/*---------------------------------------------------------------------------
 * Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.onewire.container;

// imports
import java.util.Vector;
import java.util.Enumeration;
import com.dalsemi.onewire.utils.CRC16;
import com.dalsemi.onewire.adapter.*;
import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.container.OneWireContainer;


/**
 * <P>1-Wire&#174 container for the '1K-Bit protected 1-Wire EEPROM with SHA-1
 *  Engine' family type <B>33</B> (hex), Dallas Semiconductor part number: 
 * <B>DS1961S,DS2432</B>.
 *
 * <H3> Features </H3>
 * <UL>
 *   <LI> 1128 bits of 5V EEPROM memory partitioned into four pages of 256 bits,
 *        a 64-bit write-only secret and up to 5 general purpose read/write 
 *        registers.
 *   <LI> On-chip 512-bit SHA-1 engine to compute 160-bit Message Authentication
 *        Codes (MAC) and to generate secrets.
 *   <LI> Write access requires knowledge of the secret and the capability of 
 *        computing and transmitting a 160-bit MAC as authorization.
 *   <LI> Secret and data memory can be write-protected (all or page 0 only) or 
 *        put in EPROM-emulation mode ("write to 0", page1)
 *   <LI> unique, fatory-lasered and tested 64-bit registration number (8-bit
 *        family code + 48-bit serial number + 8-bit CRC tester) assures
 *        absolute traceablity because no two parts are alike.
 *   <LI> Built-in multidrop controller ensures compatibility with other 1-Wire
 *        net products.
 *   <LI> Reduces control, address, data and power to a single data pin.
 *   <LI> Directly connects to a single port pin of a microprocessor and 
 *        communicates at up to 16.3k bits per second.
 *   <LI> Overdrive mode boosts communication speed to 142k bits per second.
 *   <LI> 8-bit family code specifies DS2432 communication requirements to reader.
 *   <LI> Presence detector acknowledges when reader first applies voltage.
 *   <LI> Low cost 6-lead TSOC surface mount package, or solder-bumped chip scale
 *        package.
 *   <LI> Reads and writes over a wide voltage range of 2.8V to 5.25V from -40C
 *        to +85C.
 * </UL>
 *
 * <P> The memory can also be accessed through the objects that are returned
 * from the {@link #getMemoryBanks() getMemoryBanks} method. </P>
 * 
 * The following is a list of the MemoryBank instances that are returned: 
 *
 * <UL>
 *   <LI> <B> Page Zero with write protection</B>
 *      <UL> 
 *         <LI> <I> Implements </I> {@link com.dalsemi.onewire.container.MemoryBank MemoryBank}, 
 *                  {@link com.dalsemi.onewire.container.PagedMemoryBank PagedMemoryBank}
 *         <LI> <I> Size </I> 32 starting at physical address 0
 *         <LI> <I> Features</I> Read/Write general-purpose non-volatile
 *         <LI> <I> Page</I> 1 page of length 32 bytes giving 29 bytes Packet data payload
 *         <LI> <I> Page Features </I> page-device-CRC and write protection.
 *      </UL> 
 *   <LI> <B> Page One with EPROM mode and write protection </B>
 *      <UL> 
 *         <LI> <I> Implements </I> {@link com.dalsemi.onewire.container.MemoryBank MemoryBank}, 
 *                  {@link com.dalsemi.onewire.container.PagedMemoryBank PagedMemoryBank} 
 *         <LI> <I> Size </I> 32 starting at physical address 32
 *         <LI> <I> Features</I> Read/Write general-purpose non-volatile
 *         <LI> <I> Page</I> 1 page of length 32 bytes giving 29 bytes Packet data payload
 *         <LI> <I> Page Features </I> page-device-CRC, EPROM mode and write protection.
 *      </UL> 
 *   <LI> <B> Page Two and Three with write protection </B>
 *      <UL> 
 *         <LI> <I> Implements </I> {@link com.dalsemi.onewire.container.MemoryBank MemoryBank}, 
 *                  {@link com.dalsemi.onewire.container.PagedMemoryBank PagedMemoryBank}
 *         <LI> <I> Size </I> 64 starting at physical address 64
 *         <LI> <I> Features</I> Read/Write general-purpose non-volatile
 *         <LI> <I> Pages</I> 2 pages of length 32 bytes giving 29 bytes Packet data payload
 *         <LI> <I> Page Features </I> page-device-CRC and write protection.
 *      </UL> 
 *   <LI> <B> Status Page that contains the secret and the status. </B>
 *      <UL>
 *         <LI> <I> Implements </I> {@link com.dalsemi.onewire.container.MemoryBank MemoryBank}, 
 *                  {@link com.dalsemi.onewire.container.PagedMemoryBank PagedMemoryBank}
 *         <LI> <I> Size </I> 24 starting at physical address 128
 *         <LI> <I> Page Features </I> Contains secret and status for the iButton.
 *      </UL>
 * </UL>
 *
 * <DD> <H4> Example 1</H4> 
 * Display some features of isMACValid where owd is an instanceof OneWireContainer33 and 
 * bank is an instanceof PagedMemoryBank: 
 * <PRE> <CODE>
 *  byte[] read_buf  = new byte [bank.getPageLength()];
 *  byte[] extra_buf = new byte [bank.getExtraInfoLength()];
 *  byte[] challenge = new byte [8];
 *
 *  // read a page (use the most verbose and secure method)
 *  if (bank.hasPageAutoCRC())
 *  {
 *     System.out.println("Using device generated CRC");
 *
 *     if (bank.hasExtraInfo())
 *     {
 *        bank.readPageCRC(pg, false, read_buf, 0, extra_buf);
 *
 *        owd.getChallenge(challenge);
 *        owd.getSecret(secret);
 *        sernum = owd.getAddress();
 *        macvalid = owd.isMACValid(bank.getStartPhysicalAddress()+pg*bank.getPageLength(),
 *                                  sernum,read_buf,extra_buf,challenge,secret);
 *     }
 *     else
 *        bank.readPageCRC(pg, false, read_buf, 0);
 *  }
 *  else
 *  {
 *     if (bank.hasExtraInfo())
 *        bank.readPage(pg, false, read_buf, 0, extra_buf);
 *     else
 *        bank.readPage(pg, false, read_buf, 0);
 *  }
 * </CODE> </PRE>
 * 
 *
 * @version 	0.00, 19 Dec 2000
 * @author 	DS
 */

public class OneWireContainer33
   extends OneWireContainer
{
   //--------
   //-------- Static Final Variables
   //--------

   /** Private Secret */
   private byte[] secret 	= new byte [8];

   /** Challenge to use for the Read Authenticate Methods */
   private byte[] challenge = new byte [8];

   /** The different memory banks for the container. */
   private MemoryBankSHAEE page1,page2,page3and4,memstatus;


   /**
    * Flag to indicate if the secret has been set.
    */
   protected boolean secretNOTset;

   /**
    * Flag to indicate if the secret is write protected.
    */
   protected boolean secretProtected;

   /**
    * Flag to indicate if the adapter has been specified.
    */
   protected boolean setAdapter;

   /**
    * Flag to indicate if the status has been checked.
    */
   protected boolean container_check;

   /**
    * block of 0xFF's used for faster read pre-fill of 1-Wire blocks
    */
   static protected byte[] ffBlock = {(byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF,
                                      (byte)0xFF,(byte)0xFF,(byte)0xFF,(byte)0xFF};

   /**
    * Constructor OneWireContainer33
    *
    *
    */
   public OneWireContainer33 ()
   {
      super();

      System.arraycopy(ffBlock,0,secret,0,8);
      System.arraycopy(ffBlock,0,challenge,0,8);

      setAdapter = false;
      container_check = false;
   }

   /**
    * Create a container with a provided adapter object
    * and the address of the iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public OneWireContainer33 (DSPortAdapter sourceAdapter, byte[] newAddress)
   {
      super(sourceAdapter, newAddress);

      System.arraycopy(ffBlock,0,secret,0,8);
      System.arraycopy(ffBlock,0,challenge,0,8);

      setAdapter = true;
      container_check = false;
      initmem();
   }

   /**
    * Create a container with a provided adapter object
    * and the address of the iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public OneWireContainer33 (DSPortAdapter sourceAdapter, long newAddress)
   {
      super(sourceAdapter, newAddress);

      System.arraycopy(ffBlock,0,secret,0,8);
      System.arraycopy(ffBlock,0,challenge,0,8);

      setAdapter = true;
      container_check = false;
      initmem();
   }

   /**
    * Create a container with a provided adapter object
    * and the address of the iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public OneWireContainer33 (DSPortAdapter sourceAdapter, String newAddress)
   {
      super(sourceAdapter, newAddress);

      System.arraycopy(ffBlock,0,secret,0,8);
      System.arraycopy(ffBlock,0,challenge,0,8);

      setAdapter = true;
      container_check = false;
      initmem();
   }

   //--------
   //-------- Methods
   //--------

   /**
    * Tells weather an adapter has been set.
    *
    * @return boolean telling weather an adapter has been set.
    */
   protected boolean adapterSet()
   {
      return setAdapter;
   }

   /**
    * Provide this container the adapter object used to access this device
    * and provide the address of this iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public void setupContainer (DSPortAdapter sourceAdapter, byte[] newAddress)
   {
      super.setupContainer(sourceAdapter,newAddress);

      if(!setAdapter)
         initmem();

      setAdapter = true;
   }

   /**
    * Provide this container the adapter object used to access this device
    * and provide the address of this iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public void setupContainer (DSPortAdapter sourceAdapter, long newAddress)
   {
      super.setupContainer(sourceAdapter,newAddress);

      if(!setAdapter)
         initmem();

      setAdapter = true;
   }

   /**
    * Provide this container the adapter object used to access this device
    * and provide the address of this iButton or 1-Wire device.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this iButton.
    * @param  newAddress        address of this 1-Wire device
    */
   public void setupContainer (DSPortAdapter sourceAdapter, String newAddress)
   {
      super.setupContainer(sourceAdapter,newAddress);

      if(!setAdapter)
         initmem();

      setAdapter = true;
   }

   /**
    * Retrieve the Dallas Semiconductor part number of the iButton
    * as a string.  For example 'DS1992'.
    *
    * @return string represetation of the iButton name.
    */
   public String getName ()
   {
      return "DS1961S";
   }

   /**
    * Retrieve the alternate Dallas Semiconductor part numbers or names.
    * A 'family' of MicroLAN devices may have more than one part number
    * depending on packaging.
    *
    * @return  the alternate names for this iButton or 1-Wire device
    */
   public String getAlternateNames ()
   {
      return "DS2432";
   }

   /**
    * Retrieve a short description of the function of the iButton type.
    *
    * @return string represetation of the function description.
    */
   public String getDescription ()
   {
      return "1K-Bit protected 1-Wire EEPROM with SHA-1 Engine.";
   }

   /**
    * Returns the maximum speed this iButton can communicate at.
    *
    * @return  max. communication speed.
    */
   public int getMaxSpeed ()
   {
      return DSPortAdapter.SPEED_OVERDRIVE;
   }

   /**
    * Get an enumeration of memory bank instances that implement one or more
    * of the following interfaces:
    * {@link com.dalsemi.onewire.container.MemoryBank MemoryBank}, 
    * {@link com.dalsemi.onewire.container.PagedMemoryBank PagedMemoryBank}, 
    * and {@link com.dalsemi.onewire.container.OTPMemoryBank OTPMemoryBank}. 
    * @return <CODE>Enumeration</CODE> of memory banks 
    */
   public Enumeration getMemoryBanks ()
   {
      Vector bank_vector = new Vector(4);

      bank_vector.addElement(page1);
      bank_vector.addElement(page2);
      bank_vector.addElement(page3and4);
      bank_vector.addElement(memstatus);

      return bank_vector.elements();
   }

   /**
    * Sets the bus master secret for this DS2432.
    *
    * @param secretset		Secret for this DS2432.
    */
   public void setContainerSecret(byte[] secretset)
   {
      System.arraycopy(secretset,0,secret,0,8);
      secretNOTset = false;
   }

   /**
    * Sets the challenge for the Read Authenticate Page
    *
    * @param challengeset  Challenge for all the memory banks.
    */
   public void setChallenge(byte [] challengeset)
   {
      System.arraycopy(challengeset,0,challenge,0,8);
   }


   /**
    * Get the secret of this device as an array of bytes.
    *
    * @param get  array of bytes containing the iButton secret
    */
   public void getSecret (byte[] get)
   {
      System.arraycopy(secret,0,get,0,8);
   }

   /**
    * Get the challenge of this device as an array of bytes.
    *
    * @param get  array of bytes containing the iButton challenge
    */
   public void getChallenge(byte[] get)
   {
      System.arraycopy(challenge,0,get,0,8);
   }

   /**
    * Get the current status of the secret.
    *
    * @return  boolean telling if the secret is set
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean isContainerSecretSet()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      return (!secretNOTset);
   }

   /**
    * Get the status of the secret, if it is write protected.
    * 
    * @return  boolean telling if the secret is write protected.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean isSecretWriteProtected()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      return secretProtected;
   }      

   /** 
    * Get the status of page zero, if it is in EPROM mode.
    *
    * @return  boolean telling if page zero is in EPROM mode.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean isPageOneEPROMmode()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      return page2.writeOnce;
   }

   /**
    * Get the status of all the pages, if they are write protected.
    *
    * @return  boolean telling if all the pages are write protected.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean isWriteProtectAllSet()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      return page3and4.readOnly;
   }

   /**
    * Get the status of page zero, if it is write protected.
    *
    * @return  boolean telling if page zero is write protected.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean isWriteProtectPageZeroSet()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      return page1.readOnly;
   }

   /** 
    *  Write protects the secret for the DS2432.
    */
   public void writeProtectSecret()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      byte[] stat = new byte[1];

      stat [0] = ( byte ) 0xAA;

      memstatus.write(8,stat,0,1);

      secretProtected = true;
   }

   /**
    * Write protect pages 0 to 3
    */
   public void writeProtectAll()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      byte[] stat = new byte[1];

      stat [0] = ( byte ) 0xAA;

      memstatus.write(9,stat,0,1);

      page1.writeprotect();
      page2.writeprotect();
      page3and4.writeprotect();
   }

   /**
    * Sets the DS2432 in EPROM mode for page 1.
    */
   public void setEPROMModePageOne()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      byte[] stat = new byte[1];

      stat [0] = ( byte ) 0xAA;

      memstatus.write(12,stat,0,1);

      page2.setEPROM();
   }

   /**
    * Write protect page 0 only.
    */
   public void writeProtectPageZero()
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      byte[] stat = new byte[1];

      stat [0] = ( byte ) 0xAA;

      memstatus.write(13,stat,0,1);

      page1.writeprotect();
   }

   /**
    * Compute Next Secret
    *
    * @param addr          address of page to use for the next secret computation.
    * @param nextsecret		new secret for this DS2432.
    * @param parialsecret  the data to put into the scrathpad in computing next secret.
    */
   public void computeNextSecret(int addr, byte[] nextsecret, byte[] partialsecret, byte[] mac)
      throws OneWireIOException, OneWireException
   {
      if(!container_check)
         container_check = this.checkStatus();

      if( (addr >= 0) && (addr < 32) )
         page1.nextSecret(addr,mac,partialsecret);
      else if( (addr >= 32) && (addr < 64) )
         page2.nextSecret(addr-32,mac,partialsecret);
      else if( (addr >= 64) && (addr < 128) )
         page3and4.nextSecret(addr-64,mac,partialsecret);
      else
         page1.nextSecret(addr,mac,partialsecret);

      System.arraycopy(secret,0,nextsecret,0,8);
   }

   /**
    * Load First Secret
    *
    * @return              boolean saying if first secret was loaded
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean loadFirstSecret(byte[] data)
      throws OneWireIOException, OneWireException
   {  
      if(!container_check)
         container_check = this.checkStatus();

      return memstatus.LoadFirSecret(0,data);
   }

   /**
    * To check the status of the part.
    *
    * @return              boolean saying the part has been checked or was checked.
    */
   protected boolean checkStatus()
      throws OneWireIOException, OneWireException
   {
      byte[] mem = new byte [8];

      if(!container_check)
      {
         memstatus.read(8,false,mem,0,8);

         if((mem [0] == ( byte ) 0xAA) || (mem [0] == ( byte ) 0x55)) 
            secretProtected = true;
         else
            secretProtected = false;

         if( ((mem [1] == ( byte ) 0xAA) || (mem [1] == ( byte ) 0x55)) ||
             ((mem [5] == ( byte ) 0xAA) || (mem [5] == ( byte ) 0x55)) )
         {
            page1.readWrite = false;
            page1.readOnly = true;
         }
         else
         {
            page1.readWrite = true; 
            page1.readOnly = false;
         }


         if((mem [4] == ( byte ) 0xAA) || (mem [4] == ( byte ) 0x55))
            page2.writeOnce = true;
         else
            page2.writeOnce = false;

         if((mem [5] == ( byte ) 0xAA) || (mem [5] == ( byte )0x55))
         {
            page2.readWrite = false;
            page2.readOnly = true;
         }
         else
         {
            page2.readWrite = true;
            page2.readOnly = false;
         }

         if((mem [5] == ( byte ) 0xAA) || (mem[5] == ( byte ) 0x55))
         {  
            page3and4.readWrite = false;
            page3and4.readOnly = true;
         }
         else
         {
            page3and4.readWrite = true;
            page3and4.readOnly = false;
         }

         memstatus.checked = true;
         page1.checked     = true;
         page2.checked     = true;
         page3and4.checked = true;
         container_check   = true;
      }
      
      return container_check;
   }

   /**
    * Initialize the memory banks and data associated with each.
    */
   private void initmem()
   {
      secretNOTset = true;

      memstatus = new MemoryBankSHAEE(this);

      // Set memory bank variables for the status and secret page
      memstatus.bankDescription = "Status Page that contains the secret and the status.";
      memstatus.generalPurposeMemory = true;
      memstatus.startPhysicalAddress = 128;
      memstatus.size = 32;
      memstatus.numberPages = 1;
      memstatus.pageLength = 32;
      memstatus.maxPacketDataLength = 32-3;
      memstatus.extraInfo = false;
      memstatus.extraInfoLength = 20;
      memstatus.readWrite = false;
      memstatus.writeOnce = false;
      memstatus.pageCRC = false;
      memstatus.readOnly = false;
      memstatus.checked = false;


      page1 = new MemoryBankSHAEE(this);

      // Set memory bank variables
      page1.bankDescription = "Page Zero with write protection.";
      page1.generalPurposeMemory = true;
      page1.startPhysicalAddress = 0;
      page1.size = 32;
      page1.numberPages = 1;
      page1.pageLength = 32;
      page1.maxPacketDataLength = 32 - 3;
      page1.extraInfo = true;
      page1.extraInfoLength = 20;
      page1.writeOnce = false;
      page1.pageCRC = true;
      page1.checked = false;


      page2 = new MemoryBankSHAEE(this);

      // Set memory bank varialbes
      page2.bankDescription = "Page One with EPROM mode and write protection.";
      page2.generalPurposeMemory = true;
      page2.startPhysicalAddress = 32;
      page2.size = 32;
      page2.numberPages = 1;
      page2.pageLength = 32;
      page2.maxPacketDataLength = 32 - 3;
      page2.extraInfo = true;
      page2.extraInfoLength = 20;
      page2.pageCRC = true;
      page2.checked = false;


      page3and4 = new MemoryBankSHAEE(this);

      // Set memory bank varialbes
      page3and4.bankDescription = "Page Two and Three with write protection.";
      page3and4.generalPurposeMemory = true;
      page3and4.startPhysicalAddress = 64;
      page3and4.size = 64;
      page3and4.numberPages = 2;
      page3and4.pageLength = 32;
      page3and4.maxPacketDataLength = 32 - 3;
      page3and4.extraInfo = true;
      page3and4.extraInfoLength = 20;
      page3and4.writeOnce = false;
      page3and4.pageCRC = true;
      page3and4.checked = false;

   }

   /**
    *  Authenticates page data given a MAC.
    *
    * @param addr          address of the data to be read
    * @param memory        the memory read from the page
    * @param mac           the MAC calculated for this function given back as the extra info
    * @param challenge     the 8 bytes of the scratch pad used in calculating the mac
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   static public boolean isMACValid(int addr, byte[] SerNum, byte[] memory, byte[] mac, 
                                    byte[] challenge, byte[] secret)
      throws OneWireException, OneWireIOException
   {
      byte[] MT  = new byte [64];
      int tmpadd = 0;
      int Temp, send_cnt, i;
      int A=0,B=0,C=0,D=0,E=0;

      System.arraycopy(secret,0,MT,0,4);

      System.arraycopy(memory,0,MT,4,32);

      System.arraycopy(ffBlock,0,MT,36,4);

      MT[40] = (byte)   (0x40 | 
                        (((addr) << 3)  & 0x08) | 
                        (((addr) >>> 5) & 0x07));

      System.arraycopy(SerNum,0,MT,41,7);

      System.arraycopy(secret,4,MT,48,4);

      System.arraycopy(challenge,4,MT,52,3);

      MT[55] = ( byte ) 0x80;

      for(i=56;i<62;i++)
         MT[i] = ( byte ) 0x00;

      MT[62] = ( byte ) 0x01;
      MT[63] = ( byte ) 0xB8;

      int[] AtoE = new int [5]; 
      ComputeSHA(MT,AtoE);
      A = AtoE[0];
      B = AtoE[1];
      C = AtoE[2];
      D = AtoE[3];
      E = AtoE[4];

      send_cnt = 0;
      Temp = E;
      for(i=0;i<4;i++)
      {
         if(mac[send_cnt++] != ( byte ) (Temp & 0x000000FF))
            return false;
         Temp >>>= 8;
      }
         
      Temp = D;
      for(i=0;i<4;i++)
      {
         if(mac[send_cnt++] != ( byte ) (Temp & 0x000000FF))
            return false;
         Temp >>>= 8;
      }

      Temp = C;
      for(i=0;i<4;i++)
      {
         if(mac[send_cnt++] != ( byte ) (Temp & 0x000000FF))
            return false;
         Temp >>>= 8;
      }

      Temp = B;
      for(i=0;i<4;i++)
      {
         if(mac[send_cnt++] != ( byte ) (Temp & 0x000000FF))
            return false;
         Temp >>>= 8;
      }

      Temp = A;
      for(i=0;i<4;i++)
      {
         if(mac[send_cnt++] != ( byte ) (Temp & 0x000000FF))
            return false;
         Temp >>>= 8;
      }

      return true;
   }

   /**
    *  To compute the SHA MAC for the DS2432.
    *
    * @param  MT            array of data to calculate MAC
    * @param  A             part of the 160 bit MAC
    * @param  B             part of the 160 bit MAC
    * @param  C             part of the 160 bit MAC
    * @param  D             part of the 160 bit MAC
    * @param  E             part of the 160 bit MAC
    */
   static protected void ComputeSHA(byte[] MT, int[] AtoE)
   {
      int[] MTword = new int[80];  //long
      int i;
      int ShftTmp;  //long
      int Temp;     //long

      for(i=0;i<16;i++)
         MTword[i] = ((MT[i*4]&0x00FF) << 24) | ((MT[i*4+1]&0x00FF) << 16) | 
                     ((MT[i*4+2]&0x00FF) << 8) | (MT[i*4+3]&0x00FF);  

      for(i=16;i<80;i++)
      {
         ShftTmp = MTword[i-3] ^ MTword[i-8] ^ MTword[i-14] ^ MTword[i-16];
         MTword[i] = ((ShftTmp << 1) & 0xfffffffe) | 
                     ((ShftTmp >>> 31) & 0x00000001);
      }

      AtoE[0] = 0x67452301;
      AtoE[1] = 0xefcdab89;
      AtoE[2] = 0x98badcfe;
      AtoE[3] = 0x10325476;
      AtoE[4] = 0xc3d2e1f0;

      for(i=0;i<80;i++)
      {
         ShftTmp = ((AtoE[0] << 5) & 0xffffffe0) | ((AtoE[0] >>> 27) & 0x0000001f);
         Temp = NLF(AtoE[1],AtoE[2],AtoE[3],i) + AtoE[4] + KTN(i) + MTword[i] + ShftTmp;
         AtoE[4] = AtoE[3];
         AtoE[3] = AtoE[2];
         AtoE[2] = ((AtoE[1] << 30) & 0xc0000000) | ((AtoE[1] >>> 2) & 0x3fffffff);
         AtoE[1] = AtoE[0];
         AtoE[0] = Temp;
      }
   }

   /**
    *  Calculation used for the DS2432 SHA MAC.
    *
    * @param  i             where in the MAC calculation
    */
   static protected int KTN (int n)
   {
      if(n<20)
         return 0x5a827999;
      else if (n<40)
         return 0x6ed9eba1;
      else if (n<60)
         return 0x8f1bbcdc;
      else
         return 0xca62c1d6;
   }      

   /**
    *  Calculation used for the DS2432 SHA MAC.
    *
    * @param  B             part of the 160 bit MAC
    * @param  C             part of the 160 bit MAC
    * @param  D             part of the 160 bit MAC
    * @param  i             where in the MAC calculation
    */
   static protected int NLF (int B, int C, int D, int n)
   {
      if(n<20)
         return ((B&C)|((~B)&D));
      else if(n<40)
         return (B^C^D);
      else if(n<60)
         return ((B&C)|(B&D)|(C&D));
      else
         return (B^C^D);
   }      

}
