
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.onewire.container;

import com.dalsemi.onewire.*;
import com.dalsemi.onewire.adapter.*;
import java.util.Date;
import com.dalsemi.onewire.utils.CRC16;


/**
 * <p>High level class implementing transactions and authentications for a DS1963S, Sha iButton.</p>
 * 
 * <p>This class makes use of several performance enhancements for TINI.
 * For instance, most methods are <code>synchronized</code> to access instance variable
 * byte arrays rather than creating new byte arrays every time a transaction
 * is performed.  This could hurt performance in multi-threaded
 * applications, but the usefulness of having several threads contending
 * to talk to a single iButton is questionable since the methods in <code>com.dalsemi.onewire.adapter.DSPortAdapter</code>
 * <code>beginExclusive(boolean)</code> and <code>endExclusive()</code> should be used.</p>
 * 
 * <p>A DS1963S SHA iButton can be a Coprocessor or a User.  A Coprocessor iButton verifies
 * signatures and signs data for User iButtons.  A Coprocessor might be located
 * inside a vending machine, where a person would bring their User iButton.  When
 * the User iButton is pressed to the Blue Dot to perform a transaction, the Coprocessor
 * would first verify that this button belongs to the system, i.e. that it knows the same
 * authentication secret (example: a Visa terminal making sure the iButton had a Visa account installed).
 * Then the Coprocessor would verify the signed data, probably money, to make sure it was
 * valid.  If someone tried to overwrite the money file, even with a previously valid
 * money file (an attempt to 'restore' a previous amount of money), the signed file would 
 * be invalid because the signature includes the write cycle counter, which is incremented 
 * every time a page is written to.  The write cycle counter is read-only and does not roll over, 
 * so the previous amount of money could not be restored by rolling the write counter.
 * The Coprocessor verifies the money, then signs a new data file that
 * contains the new amount of money. </p>
 * 
 * <p>There are two secrets involved with the transaction process.  The first secret is
 * the authentication secret.  It is used to validate a User iButton to a system.  The Coprocessor
 * iButton has the system authentication secret installed.  On User iButtons, the 
 * system authentication secret is merged with binding data and the unique address 
 * of the User iButton to create a unique device authentication secret.  The second secret
 * is a signing secret.  This secret only exists on the Coprocessor iButton, and is used
 * to sign and verify data such as money.  These secrets are inaccessible outside the
 * iButton.  Once they are installed, they cannot be retrieved.</p>
 *
 * <h3>Usage</h3>
 * <ol>
 *     <li><a href="#initcopr">Initializing a Coprocessor iButton</a></li>
 *     <li><a href="#initrov">Initializing a User iButton</a>
 *         <ol>
 *             <li><A href="#i">Initializing a <code>SHAiButton</code> to 
 *                              act as a Coprocessor</a></li>
 *             <li><A href="#ii">Using a Coprocessor to initialize a User iButton</a></li>
 *         </ol>
 *     </li>
 *     <li><a href="#debit">Performing an Authentication and a Transaction</a></li>
 * </ol>
 *
 * <a name="initcopr"></a>
 * <b>Initializing a Coprocessor iButton</b><br>
 * <p>
 * There are two main operations involved in initializing a DS1963S to be 
 * a Coprocessor iButton for transactions.  First, it must have a Coprocessor
 * file that defines parameters for the system.  Second, it must install a master
 * authentication secret and a signing secret.</p>
 *
 * <p>Dallas Semiconductor has created a suggested format for Coprocessor
 * files to help developers implement systems for authentication and transactions.
 * The structure is as follows:
 * <dl><dd><table>
 *    <tr>
 *        <td> <center>Field</center> </td>
 *        <td> # bytes </td>
 *        <td> <center>Sample data </center></td>
 *    </tr>
 *    <tr>
 *        <td>Name of user account data file</td>
 *        <td> <center>5</center> </td>
 *        <td> <center>{'D','L','S','M',102}</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Signing page number </td>
 *        <td> <center>1</center> </td>
 *        <td> <center>8</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Authentication page number </td>
 *        <td> <center>1</center> </td>
 *        <td> <center>7 </center></td>
 *    </tr>
 *    <tr>
 *        <td> Workspace page number</td>
 *        <td><center> 1</center> </td>
 *        <td> <center>9 </center></td>
 *    </tr>
 *    <tr>
 *        <td> Version number </td>
 *        <td> <center>1</center> </td>
 *        <td> <center>1</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Installation data code </td>
 *        <td> <center>4 </center></td>
 *        <td> <center>10182000</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Binding data </td>
 *        <td> <center>32</center> </td>
 *        <td> <center>32 0xff bytes</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Binding code </td>
 *        <td> <center>7</center> </td>
 *        <td> <center>7 0xff bytes</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Signing challenge</td>
 *        <td> <center>3</center> </td>
 *        <td> <center>3 0x00 bytes</center></td>
 *    </tr>
 *    <tr>
 *        <td> Provider name length (<i>LN</i>)</td>
 *        <td> <center>1</center> </td>
 *        <td> <center>20</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Initial Signature length (<i>LS</i>)</td>
 *        <td> <center>1</center> </td>
 *        <td> <center>20</center> </td>  
 *    </tr>
 *    <tr>
 *        <td> Extra data length (<i>LE</i>)</td>
 *        <td> <center>1</center> </td>
 *        <td> <center>0</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Service provider name </td>
 *        <td> <center><i>LN</i></center> </td>
 *        <td> <center>"Dallas Semiconductor"</center> </td>
 *    </tr>
 *    <tr>
 *        <td> Initial signature </td>
 *        <td> <center><i>LS</i></center> </td>
 *        <td> <center>20 0x00 bytes </center></td>
 *    </tr>
 *    <tr>
 *        <td> Extra data </td>
 *        <td> <center><i>LE</i></center> </td>
 *        <td> </td>
 *    </tr>
 *    <tr>
 *        <td> Encryption algorithm code </td>
 *        <td> <center>1</center> </td>
 *        <td> <center>0</center> </td>
 *    </tr>
 * </table></dd></dl>
 * </p>
 *
 *
 * <p>This class assumes the use of TMEX files.  The code below shows how to set up
 * a Coprocessor iButton using the TMEX file format and the service file format
 * discussed above.</p>
 *
 * <pre><code>
 *       <i>//preformatted root directory page for TMEX</i>
 *       byte[] page0 = 
 *       {
 *           (byte)0x0F, (byte)0xAA, (byte)0x00, (byte)0x80, (byte)0x03, (byte)0x00, (byte)0x00, (byte)0x00, 
 *           (byte)0x43, (byte)0x4F, (byte)0x50, (byte)0x52, (byte)0x00, (byte)0x01, (byte)0x01, (byte)0x00,
 *           (byte)0x74, (byte)0x9C, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, 
 *           (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF
 *       };
 *
 *       . . . 
 *
 *       <i>//adapter is a DSPortAdapter that has just found a DS1963S</i>
 *       OneWireContainer18 owc18 = new OneWireContainer18();
 *       owc18.setupContainer(adapter,adapter.getAddressAsLong());
 *
 *       <i>//we will control the speed of the 1-Wire bus, don't do speed check</i>
 *       owc18.setSpeedCheck(false);
 *       SHAiButton sha = new SHAiButton(owc18);
 *       
 *       <i>//first we need to accumulate data for our service file</i>
 *       ByteArrayOutputStream baos = new ByteArrayOutputStream();
 *       baos.write("DLSM".getBytes());                 <i>//coprocessor service file name</i>
 *       baos.write(102);                               <i>//coprocessor service file extension</i>
 *       baos.write(8);                                 <i>//signing page</i>
 *       baos.write(7);                                 <i>//authentication page</i>
 *       baos.write(9);                                 <i>//workspace page</i>
 *       baos.write(1);                                 <i>//version number</i>
 *       baos.write(10);                                <i>//date stamp: month</i>
 *       baos.write(18);                                <i>//date stamp: date</i>
 *       baos.write(20);                                <i>//date stamp: year / 100</i>
 *       baos.write(0);                                 <i>//date stamp: year % 100</i>
 *       <i>//32 bytes of bind data</i>
 *       baos.write("01234567890123456789012345678901".getBytes()); 
 *       baos.write("binding".getBytes());              <i>//bind code, 7 bytes</i>
 *       baos.write(new byte[3]);                       <i>//signing challenge, 3 bytes</i>
 *       baos.write("Dallas Semiconductor".length());   <i>//provider of the service</i>
 *       baos.write(20);                                <i>//initial signature length (20)</i>
 *       baos.write(0);                                 <i>//extra data length</i>
 *       baos.write("Dallas Semiconductor".getBytes()); <i>//provider of the service</i>
 *       baos.write(new byte[20]);                      <i>//initial signature</i>
 *       baos.write("".getBytes());                     <i>//any extra data</i>
 *       baos.write(0);                                 <i>//an encryption code</i>
 *       byte[] data = baos.toByteArray();
 *        
 *       <i>//start writing the page at page 1 of the Coprocessor,</i>
 *       <i>//since page 0 will be taken by the TMEX root directory file</i>
 *       int page = 1;
 *       byte[] this_page = new byte[32];
 *       int index = 0;
 *       int length;
 *       while (index &lt; data.length)
 *       {
 *           <i>//this loop handles breaking the data into</i> 
 *           <i>//TMEX file structure compatible pages</i>
 *           if (data.length - index &gt; 28) 
 *               length = 28;
 *           else 
 *               length = data.length - index;
 *
 *           this_page[0] = (byte)(length + 1);  
 *           System.arraycopy(data,index,this_page,1,length);
 *           index += length;
 *           if (index<data.length)
 *               this_page[length+1] = (byte)(page + 1);
 *           else
 *               this_page[length+1] = (byte) 0;   //last page
 *               
 *           int crc = CRC16.compute(this_page,0,length+2,page);
 *           <i>//make sure this is an inverted CRC</i>
 *           crc = (~crc) & 0x0ffff;
 *             
 *           this_page[length+2] = (byte) crc;
 *           this_page[length+3] = (byte) (crc &gt;&gt; 8);
 *              
 *           owc18.writeDataPage(page,this_page);
 *           page++;
 *       }
 *       <i>//now we have written the entire COPR.0 file, now we </i>
 *       <i>//need to fix up page 0, the root directory</i>
 *
 *       <i>//make a bitmask of which pages we have used.  since we</i>
 *       <i>//have used all pages from 0 to (page-1), this trick will work</i>
 *       int mask = (0x01 &lt;&lt; (page)) - 1;
 *           
 *       <i>//now put that int (LSB first) in page0[4..7]</i>
 *       page0[4] = (byte)(mask);
 *       page0[5] = (byte)(mask &gt;&gt; 8);
 *       page0[6] = (byte)(mask &gt;&gt; 16);
 *       page0[7] = (byte)(mask &gt;&gt; 24 );
 *           
 *       <i>//now how many pages were used for file COPR.000?</i>
 *       <i>//if we used (page) pages, and page 0 is the root, then there are page-1 pages in the file</i>
 *       <i>//stick it in the file at position 14</i>
 *       page0[14] = (byte)(page-1);
 *           
 *       <i>//now recalculate the CRC for the root directory</i>
 *       int CRC = ~CRC16.compute(page0,0,page0[0]+1);
 *       page0[page0[0]+1] = (byte)CRC;
 *       page0[page0[0]+2] = (byte)(CRC &gt;&gt; 8);
 *
 *       //now we have to fix page 0
 *       if ((!owc18.eraseScratchPad(0)) || (!owc18.writeDataPage(0,page0)))
 *       {
 *           System.out.println("Failed to write page 0");
 *           return;
 *       }
 *
 *       <i>//done with root directory, now install signing secret and an authentication secret</i>
 *       byte[] sign_secret = "this is my signing secret".getBytes();
 *       byte[] auth_secret = "this is my authentication secret".getBytes();
 *       
 *       if (!owc18.installMasterSecret(8,sign_secret,0))
 *           System.out.println("Could not install master signing secret!");
 *       else 
 *           System.out.println("Signing secret installed.");
 *       if (!owc18.installMasterSecret(7,auth_secret,7))
 *           System.out.println("Could not install master authentication secret!");
 *       else 
 *           System.out.println("Authentication secret installed.");
 * </code></pre>
 *
 * <p>In order to initialize a <code>SHAiButton</code> object to be used as
 * a Coprocessor iButton, read the "COPR.0" file from the DS1963S using
 * the <code>readFile(int)</code> command.  Use the file data to make calls to
 * setup this iButton as a Coprocessor.  See {@link #isCoprocessor() isCoprocessor()}
 * for which parameters must be set to act as a Coprocessor.  The next section
 * also describes setting up a <code>SHAiButton</code> object to act as a Coprocessor.</p>
 *
 * <a name="initrov"></a>
 * <b>Initializing a User iButton</b><br>
 * <p>Initializing a User iButton also requires two steps.  The first is finding
 * and gathering information from a Coprocessor iButton.  The second is installing
 * an authentication secret and an initial money page on the User iButton.</p>
 *
 * <a name="i"></a>
 * <p><b>I. </b>The following code shows how gather data from a Coprocessor iButton
 * and setup a <code>SHAiButton</code> object to act as a Coprocessor.</p>
 * <code><pre>
 *       <i>//adapter is a DSPortAdapter, ID is the byte[] address of the DS1963S</i>
 *       <i>// that has already been found</i>
 *       SHAiButton copr = null;
 *       OneWireContainer18 ds1963S = (OneWireContainer18) adapter.getDeviceContainer(ID);
 *       ds1963S.setupContainer(adapter,ID);
 *       ds1963S.setSpeedCheck(false);
 *       SHAiButton sha = new SHAiButton(ds1963S);
 *       byte[] filebuffer = new byte[400];
 *       filelength = sha.readFile(0, filebuffer);
 *
 *       <i>//look for the file "COPR.000"</i>
 *       byte[] file = "COPR".getBytes();
 *       int filepage = -1;
 *
 *       for (i=0;i &lt; filelength;i++)
 *       {
 *           if (filebuffer[i]==file[0])
 *             if (filebuffer[i+1]==file[1])
 *               if (filebuffer[i+2]==file[2])
 *                 if (filebuffer[i+3]==file[3])
 *                   if (filebuffer[i+4]==(byte)0)  <i>//000 is the extension</i>
 *                       filepage = filebuffer[i+5];
 *       }
 *       if (filepage==-1)
 *       {
 *           System.out.println("Could not find the file page!!!");
 *       }
 *       else
 *       {
 *           filelength = sha.readFile(filepage, filebuffer);
 *           sha.setFilename(filebuffer,0);
 *           sha.setSigningPageNumber(filebuffer[5]);
 *           sha.setAuthenticationPageNumber(filebuffer[6]);
 *           sha.setWorkspacePageNumber(filebuffer[7]);
 *           <i>//note: not all Dallas Semiconductor demo's include the bind</i>
 *           <i>//  data and bind code in the COPR.000 file, which is why</i>
 *           <i>//  we check the file length here</i>
 *           if (filelength &gt;= 45)
 *               sha.setBindData(filebuffer,13);
 *           if (filelength &gt;= 52)
 *               sha.setBindCode(filebuffer,45);
 *           if (sha.isCoprocessor())
 *           {
 *               copr = sha;
 *               System.out.println("Found a coprocessor for service \""+sha.getUserFileName()+"\"");
 *           }
 *       }                        
 * </pre></code>
 *
 * <a name="ii"></a>
 * <p><b>II. </b>Once the Coprocessor iButton has been set, any other DS1963S can be set as a User iButton.
 * The following code shows how to create a User iButton that would be compatible with 
 * the Coprocessor iButton created in the section above.</p>
 * <code><pre>
 *       <i>//byte[] page0 should be the same as in the Coprocessor initialization section</i>
 *       <i>//owc18 has already been set up to talk to a new DS1963S</i>
 *       owc18.setSpeedCheck(false);
 *       
 *       SHAiButton sha = new SHAiButton(owc18);
 * 
 *       int auth_page = 13;   <i>//put the money page and authentication data on page 13</i>
 *       int tid = 0x1234;     <i>//transaction ID, defined by credit/debit authority</i>
 *       int fact = 0x8b48;    <i>//ISO4217 monetary unit code and multiplier</i>
 *       int data_type = 0;    <i>//certificate type, defined by credit/debit authority</i>
 *       
 *       byte[] auth_secret = "this is my authentication secret".getBytes();
 *
 *       if (!owc18.installMasterSecret(auth_page,auth_secret,auth_page&7))
 *           System.out.println("Could not install master authentication secret!");
 *       else 
 *           System.out.println("Authentication secret installed.");
 *
 *       <i>//now we need to bind this data to the user iButton</i>
 *       if (!owc18.bindSecretToiButton(auth_page,copr.getBindData(), copr.getBindCode(), auth_page & 7))
 *       {
 *           System.out.println("Failed to bind the secret to the iButton");
 *       }
 *       else
 *           System.out.println("Authentication secret bound to iButton");
 *
 *       <i>//now the authentication secret has been installed, </i>
 *       <i>//we need to put some initial money on the page</i>
 *       byte[] default_page = new byte[32];
 *       default_page[1] = (byte)0x34;
 *       default_page[2] = (byte)0x12;
 *       default_page[3] = (byte)0x48;
 *       default_page[4] = (byte)0x8b;
 *
 *       sha.setUser(auth_page);
 *       <i>//this puts $100.00 on the iButton (10000 pennies)</i>
 *       <i>//copr is defined in part I of this User initialization example</i>
 *       if (copr.signDataFile(sha,10000,-1, default_page))
 *           System.out.println("Successfully set up user iButton");
 *       else
 *           System.out.println("Failed to set up user iButton");
 *
 *       <i>//we've put money on the iButton, now we need a root directory</i>
 *       char[] name = copr.getUserFileName().toCharArray();
 *       int ext = copr.getUserFileExtension();
 *       page0[8] = (byte)name[0];    
 *       page0[9] = (byte)name[1];    
 *       page0[10] = (byte)name[2];    
 *       page0[11] = (byte)name[3];    
 *       page0[12] = (byte)ext;    
 *       page0[13] = (byte)auth_page;
 *       page0[14] = (byte)1;
 *
 *       <i>//the mask should represent that we are using page 0 and page auth_page</i>
 *       int mask = 0x01 | (0x01 &lt;&lt; auth_page);
 *           
 *       <i>//now put that int (LSB first) in page0[4..7]</i>
 *       page0[4] = (byte)(mask);
 *       page0[5] = (byte)(mask &gt;&gt; 8);
 *       page0[6] = (byte)(mask &gt;&gt; 16);
 *       page0[7] = (byte)(mask &gt;&gt; 24);
 *      
 *       <i>//now redo the CRC and write it</i>
 *       int CRC = ~CRC16.compute(page0,0,page0[0]+1);
 *       page0[page0[0]+1] = (byte)CRC;
 *       page0[page0[0]+2] = (byte)(CRC &gt;&gt; 8);
 *
 *       <i>//now put this new page 0 on the iButton as a root directory</i>
 *       if ((!owc18.eraseScratchPad(0)) || (!owc18.writeDataPage(0,page0)))
 *       {
 *           System.out.println("Failed to write page 0");
 *           return;
 *       }   
 * </pre></code>
 *
 * <a name="debit"></a>
 * <b>Performing an Authentication and a Transaction</b><br>
 * <p>
 * The code in <code>OneWireContainer18</code> and this class has been written to 
 * make transactions fast in Java.  While the example code for initializing
 * Coprocessors and Users is not extremely fast, the code in the critical path
 * of a transaction is fast.  The critical path of a transaction can be defined as everything
 * that happens from the time an iButton is inserted for a transaction until 
 * a new data file has been signed and installed on the User iButton.  This means
 * that code to find a Coprocessor and initialize a <code>SHAiButton</code> object to
 * act as a Coprocessor is not in the critical path.  The code below uses a
 * <code>SHAiButton</code> object <code>copr</code>.  It can be set up in the same
 * manner as described in section I of "Initializing a User iButton".
 * </p>
 * <code><pre>
 *       OneWireContainer18 owc = new OneWireContainer18();
 *       owc.setSpeedCheck(false);
 *       SHAiButton user = new SHAiButton();
 *       byte[] file     = copr.getUserFileName().getBytes();
 *       byte   ext      = ( byte ) copr.getUserFileExtension();
 *
 *       <i>//adapter is our DSPortAdapter object, it has just found</i>
 *       <i>// a DS1963S to transact with</i>
 *
 *       <i>//CRITICAL PATH BEGINS HERE, let's time it</i>
 *       long    t1      = System.currentTimeMillis();
 *       <i>//first let's start talking in overdrive</i>
 *       adapter.reset();
 *       adapter.putByte(0x3c);
 *       adapter.setSpeed(DSPortAdapter.SPEED_OVERDRIVE);
 *       adapter.reset();
 *
 *       adapter.getAddress(ID);
 *       owc.setupContainer(adapter, ID);
 *       user.setiButton(owc);
 *
 *       <i>//now we need to find the file installed!</i>
 *       <i>//filebuffer is a previously defined, large byte array (~400 bytes)</i>
 *       filelength = user.readFile(0, filebuffer);
 *
 *       int filepage = -1;
 *
 *       for (i = 0; i < filelength-5; i++)
 *       {
 *          if (filebuffer [i] == file [0])
 *             if (filebuffer [i + 1] == file [1])
 *                if (filebuffer [i + 2] == file [2])
 *                   if (filebuffer [i + 3] == file [3])
 *                      if (filebuffer [i + 4] == ext)
 *                      {
 *                         <i>//we have found the service file on this iButton</i>
 *                         filepage = filebuffer [i + 5];
 *                         i = filelength;
 *                      }
 *       }
 *
 *       if (filepage == -1)
 *       {
 *          System.out.println("No service installed!");
 *       }
 *       else
 *       {
 *          <i>//set up this iButton as a User iButton</i>
 *          user.setUser(filepage);
 *          int wcc = copr.verifyAuthentication(user, pagedata);
 *
 *          if (wcc > 0)
 *          {
 *             <i>//verify that our money is valid</i>
 *             if (copr.verifyUserMoney(pagedata, user,wcc))
 *             {
 *                <i>//this is how the cash is stored in our money page</i>
 *                int cash = (pagedata [5] & 0x0ff) | ((pagedata [6] & 0x0ff) << 8)
 *                           | ((pagedata [7] & 0x0ff) << 16);
 *                <i>//amount is the amount to be deducted or added</i>
 *                if (cash + amount >= 0)
 *                {
 *                   <i>//put the new amount of money onto the iButton</i>
 *                   if (copr.signDataFile(user, cash + amount, wcc, pagedata))
 *                   {
 *                      <i>//CRITICAL PATH IS OVER, stop timing</i>
 *                      long t2      =  System.currentTimeMillis();
 *                      int  newcash = cash + amount;
 *                      <i>//now report our transaction</i>
 *
 *                      System.out.println("-----------------------");
 *                      System.out.println("| Device  : "+ Long.toHexString(t));
 *                       
 *                      <i>//note that I am assuming the money is stored as US $0.01</i>
 *                      <i>//rather than reading the monetary unit code and multiplier from </i>
 *                      <i>//the iButton</i>
 *                      System.out.println("| Balance : $ " + (newcash / 100)+ "." + (newcash % 100) + " US");
 *                      System.out.println("| Transaction time : " + (t2 - t1) + " ms");
 *                      System.out.println("-----------------------\r\n");
 *                   }
 *                }
 *             }
 *          }
 *       }
 *       <i>//put the 1-Wire bus back at normal speed</i>
 *       adapter.setSpeed(DSPortAdapter.SPEED_REGULAR);
 *       adapter.reset();
 * </pre></code>
 * <h3>Relevant Links:</h3>
 * <ul>
 *    <li>1963S Datasheet: <A HREF="http://www.dalsemi.com/datasheets/pdfs/1963S.pdf"> http://www.dalsemi.com/datasheets/pdfs/1963S.pdf</A>
 *    <li>iButton.com SHA iButton site: <A HREF="http://www.ibutton.com/ibuttons/ds1963s.html"> http://www.ibutton.com/ibuttons/ds1963s.html</A>
 *    <li>TMEX site: <A HREF="http://www.ibutton.com/software/tmex/index.html"> http://www.ibutton.com/software/tmex/index.html</A>
 * </ul>
 * <p>This class makes no calls to the adapter level, i.e. it does no direct talking to
 * the device.  It uses the methods in <code>OneWireContainer18</code> to implement
 * transactions.</p>
 *
 * @see OneWireContainer18
 * 
 * @version    0.00, 28 Aug 2000
 * @author     KLA
 */
public class SHAiButton
{
   private OneWireContainer18 ibc = null;   //access to the underlying container
   private byte[]             address;         //take out an extra method call or so
   private boolean            isCoprocessor;   //is this button a coprocessor?
   private boolean            isUser;          //is this button a user?
   private byte[]             bind_data = new byte [32];
   private byte[]             bind_code = new byte [7];

   //data that means something in a coprocessor
   private byte[] filename = null;   //filename is jsut the four bytes in user_account_file

   //plus one byte for user_account_file_extension
   private int signing_page_number        = -1;
   private int authentication_page_number = -1;
   private int workspace_page_number      = -1;

   //bind_data and bind_code already defined, is OK - they can be defined for both buttons
   private byte[] signing_challenge = new byte [3];
   private byte[] initial_signature = new byte [20];   //all 0's is legal for this one

   //data that means something in a user
   private int    user_account_file_page;
   private byte[] generic_money_page = new byte [32];
   private byte[] scratchpad_buffer = new byte[32];
   private int    last_error = 0;

   /**
    * Useful constant for <code>getLastError()</code>.  Means no error occurred.
    *
    * @see #getLastError()
    */
   public static final int NO_ERROR = 0;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during an Erase Scratch Pad activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int ERASE_SCRATCHPAD_ERROR = 1;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during an Erase Scratch Pad activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int WRITE_SCRATCHPAD_ERROR = 2;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a Read Scratch Pad activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int READ_SCRATCHPAD_ERROR = 3;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a Read Memory Page activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int READ_MEMORY_PAGE_ERROR = 4;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a Read Authenticated Memory Page activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int READ_AUTHENTICATED_ERROR = 5;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a Secret Binding activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int BIND_SECRET_ERROR = 6;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a Write Memory Page activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int WRITE_MEMORY_PAGE_ERROR = 7;

   /**
    * Useful constant for <code>getLastError()</code>.  Means an 
    * error occurred during a cryptographic activity with the DS1963S.
    * Probably due to a communication error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int SHA_FUNCTION_ERROR = 8;

   /**
    * Useful constant for <code>getLastError()</code>.  Means a 
    * CRC from the DS1963S was invalid.  Probably due to a communication 
    * error, meaning a retry might succeed.
    *
    * @see #getLastError()
    */
   public static final int CRC_ERROR = 9;

   /**
    * Useful constant for <code>getLastError()</code>.  Means 
    * that the coprocessor <code>SHAiButton</code> object has not 
    * been successfully setup.
    *
    * @see #getLastError()
    */
   public static final int NO_COPROCESSOR_ERROR = 10;

   /**
    * Useful constant for <code>getLastError()</code>.  Means 
    * that the user <code>SHAiButton</code> object has not been 
    * successfully setup.
    *
    * @see #getLastError()
    */
   public static final int NO_USER_ERROR = 11;

   /**
    * Useful constant for <code>getLastError()</code>.  It is possible this error
    * is due to a communication error, and that a retry will be
    * successful.  However, it can also mean that the user <code>SHAiButton</code>
    * does not belong to the system (could not authenticate).
    *
    * @see #getLastError()
    */
   public static final int AUTHENTICATION_FAILED_ERROR = 12;

   /**
    * Useful constant for <code>getLastError()</code>.  It is possible this error
    * is due to a communication error, and that a retry will be
    * successful.  However, it can also mean that the user <code>SHAiButton</code>'s
    * money data file is invalid (has been tampered with).
    *
    * @see #getLastError()
    */
   public static final int VERIFICATION_FAILED_ERROR = 13;




   /**
    * <p>Creates a new object to handle authentications and transactions
    * with a DS1963S SHA iButton.  This constructor does not initialize
    * any parameters, but only creates allocation for several instance 
    * variables.  Before doing any processing using this <code>SHAiButton</code>,
    * call <code>setiButton(OneWireContainer18)</code> to initialize this
    * object.  This method is meant to be used outside of the 'critical path'
    * of a transaction:
    * <code><pre>
    *     //initialize variables to use in the transaction loop here
    *     . . .
    *     SHAiButton user = new SHAiButton();
    *     OneWireContainer18 owc18 = new OneWireContainer18();
    *     . . .
    *     while (allowingTransactions)
    *     {
    *         //here begins the 'critical path' of the transaction
    *         . . . 
    *         //find a new DS1963S iButton inserted
    *         // 'ID' is the 8 byte address of the device
    *         // 'adapter' is the DSPortAdapter object to talk to the device
    *         owc18.setupContainer(adapter, ID);
    *         user.setiButton(owc18);
    *         . . .
    *         //now perform transaction
    *     }
    * </pre></code>
    * This is less expensive than creating a new <code>SHAiButton</code>
    * in every instance of the loop.
    *</p>
    *
    * @see #SHAiButton(OneWireContainer18)
    * @see #setiButton(OneWireContainer18)
    */
   public SHAiButton()
   {
      //assume you'll later call setiButton...
      
      //or else!!!
   }

   /**
    * Creates a new object to handle authentications and transactions
    * with a DS1963S SHA iButton.  See {@link #SHAiButton() the default constructor}
    * for speed considerations when creating a new <code>SHAiButton</code> object
    * for a User iButton.
    *
    * @param ibc the initialized <code>OneWireContainer</code> object for the DS1963S
    *        that will belong to this <code>SHAiButton</code>
    *
    * @see #SHAiButton()
    */
   public SHAiButton (OneWireContainer18 ibc)
   {
      this.ibc          = ibc;
      isCoprocessor     = false;
      isUser            = false;
      address           = ibc.getAddress();

      //the OneWireContainer18 already has an array of all FF's
      System.arraycopy(ibc.FF,0,bind_data,0,32);
      System.arraycopy(ibc.FF,0,bind_code,0,7);
   }
   
   /**
    * <p>Sets the <code>OneWireContainer18</code> object that this <code>SHAiButton</code>
    * will use to communicate with a DS1963S.  This method should be used in conjunction
    * with the constructor {@link #SHAiButton()}.  Note that the following two code
    * snippets are functionally equivalent:
    * <code><pre>
    *     . . . 
    *     //owc18 is a OneWireContainer18 already initialized
    *     SHAiButton sha = new SHAiButton(owc18);
    *     . . .
    * </pre></code>
    * is functionally equivalent to
    * <code><pre>
    *     . . . 
    *     //owc18 is a OneWireContainer18 already initialized
    *     SHAiButton sha = new SHAiButton();
    *     sha.setiButton(owc18);
    *     . . .
    * </pre></code>
    * The reasons for using the latter code are discussed in the <code>SHAiButton()</code>
    * documentation.</p>
    *
    * @see #SHAiButton()
    * @see #SHAiButton(OneWireContainer18)
    */
   public void setiButton(OneWireContainer18 ibc)
   {
      this.ibc          = ibc;
      isCoprocessor     = false;
      isUser            = false;

      address           = ibc.getAddress();

      System.arraycopy(ibc.FF,0,bind_data,0,32);
      System.arraycopy(ibc.FF,0,bind_code,0,7);
   }
   

   /**
    * Sets the binding data for this SHA iButton.  This should be 32 bytes.
    * Anything shorter will result in an <code>java.lang.ArrayIndexOutOfBoundsException</code>.
    * Data beyond the first 32 bytes will be ignored.  This data is used to bind secrets
    * to User iButtons and also to re-create user signatures in
    * Coprocessor iButtons.
    *
    * @param buf buffer with the binding data
    * @param offset offset to start reading from the buffer (32 bytes 
    *        will be read from <code>offset</code>)
    *
    * @see #getBindData()
    * @see #setBindCode(byte[],int)
    */
   public synchronized void setBindData (byte[] buf, int offset)
   {
      System.arraycopy(buf, offset, bind_data, 0, 32);
   }

   /**
    * Sets the binding code for this SHA iButton.  This should be 7 bytes.
    * Anything shorter will result in an <code>java.lang.ArrayIndexOutOfBoundsException</code>.
    * Data beyond the first 7 bytes will be ignored.  This data is written to
    * the scratchpad for binding secrets to User iButtons and also to recreate
    * User iButton signatures on Coprocessor iButtons.
    *
    * @param buf buffer with the binding data
    * @param offset offset to start reading from the buffer (7 bytes 
    *        will be read from <code>offset</code>)
    *
    * @see #getBindCode()
    * @see #setBindData(byte[],int)
    */
   public synchronized void setBindCode (byte[] buf, int offset)
   {
      System.arraycopy(buf, offset, bind_code, 0, 7);
   }

   /**
    * Gets the 32-byte binding data.  This data is used to bind secrets
    * to User iButtons and also to re-create user signatures in
    * Coprocessor iButtons.
    *
    * @return the binding data (length 32)
    * 
    * @see #getBindCode()
    * @see #setBindData(byte[],int)
    */
   public synchronized byte[] getBindData ()
   {
      byte[] b = new byte [bind_data.length];

      System.arraycopy(bind_data, 0, b, 0, b.length);

      return b;
   }

   /**
    * Get the 7-byte binding code.  This data is written to
    * the scratchpad for binding secrets to User iButtons and also to recreate
    * User iButton signatures on Coprocessor iButtons.
    *
    * @return the binding code (length 7)
    *
    * @see #getBindData()
    * @see #setBindCode(byte[],int)
    */
   public synchronized byte[] getBindCode ()
   {
      byte[] b = new byte [bind_code.length];

      System.arraycopy(bind_code, 0, b, 0, b.length);

      return b;
   }

   private byte[] answer_readpg        = new byte [42];
   private byte[] answer_pad_challenge = new byte [32];

   /**
    * <p>Answers a challenge from the Coprocessor iButton by signing a data page.
    * This method is used by the User iButton.</p>
    *
    * <p>To authenticate a User iButton to a system, the Coprocessor iButton 
    * generates a random challenge.  This way, an attacker cannot simply memorize
    * one successful response and fake its way into the system.  The User iButton
    * signs a data page with its system authentication secret.  The random number generated
    * by the Coprocessor and the write cycle counter of the user's data page are also
    * part of the signature.  If the Coprocessor can verify this signature,
    * then the User iButton belongs to the system.</p>
    * 
    * <p>Note that authentication does not guarantee the validity of the data 
    * in the page.  It must be verified using <code>verifyUserMoney()</code>.</p>
    *
    * <p>The only true input to this method is the random byte array 
    * <code>challenge</code>.  The other byte arrays are used to return 
    * information that the Coprocessor iButton needs to verify the signature.</p>
    *
    * @param challenge a 3 byte random challenge (generated by the Coprocessor DS1963S)
    * @param mac buffer for the returned signature generated by the 
    *        User DS1963S (must be at least 20 bytes long)
    * @param pagedata buffer for the page data the User DS1963S signed 
    *        (must be at least 32 bytes long)
    * 
    * @return the write cycle counter value of the page returned in <code>pagedata</code>,
    *         of -1 if an error occurred  (use <code>getLastError()</code> for more
    *         information on the type of error)
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    *
    * @see #verifyUserMoney(byte[],SHAiButton,int)
    * @see #generateChallenge(int,int,byte[])
    * @see #getLastError()
    */
   public synchronized int answerChallenge (byte[] challenge, byte[] mac,
                                            byte[] pagedata)
      throws OneWireIOException, OneWireException
   {
      last_error = NO_ERROR;

      System.arraycopy(challenge, 0, answer_pad_challenge, 20, 3);

      if (ibc.eraseScratchPad(user_account_file_page))
      {
         ibc.useResume(true);

         if (ibc.writeScratchPad(user_account_file_page, 0,
                                 answer_pad_challenge, 0, 32))
         {
            boolean readOK = ibc.readAuthenticatedPage(user_account_file_page,
                                                       answer_readpg, 0);
            int     wcc    = (answer_readpg [35] & 0x0ff);

            wcc = (wcc << 8) | (answer_readpg [34] & 0x0ff);
            wcc = (wcc << 8) | (answer_readpg [33] & 0x0ff);
            wcc = (wcc << 8) | (answer_readpg [32] & 0x0ff);

            int len = ibc.readScratchPad(scratchpad_buffer, 0);

            if ((!readOK) || (len < 0))
            {
               if (!readOK)
                  last_error = READ_AUTHENTICATED_ERROR;
               else
                  last_error = READ_SCRATCHPAD_ERROR;

               return -1;
            }

            System.arraycopy(answer_readpg, 0, pagedata, 0, 32);
            System.arraycopy(scratchpad_buffer, 8, mac, 0, 20);
            ibc.useResume(false);

            return wcc;
         }
         else
         {
            last_error = this.WRITE_SCRATCHPAD_ERROR;

            ibc.useResume(false);

            return -1;
         }
      }

      last_error = ERASE_SCRATCHPAD_ERROR;

      ibc.useResume(false);

      return -1;
   }

   /**
    * <p>Generates a 3 byte random challenge in the iButton, sufficient to be used
    * as a challenge to be answered by a User iButton.  This method will normally
    * be called by a Coprocessor iButton.</p>
    *
    * <p>The DS1963S will generate 20 bytes of pseudo random data, though only 
    * 3 bytes are needed for the challenge.  Programs can add more 'randomness'
    * by selecting different bytes from the 20 bytes of random data using the 
    * <code>offset</code> parameter.</p>
    *
    * <p>The random number generator is actually the DS1963S's SHA engine, which requires 
    * page data to compute a hash.  Select a page number with the <code>page_number</code>
    * parameter.</p>
    *
    * @param page_number page number that the random number generation will be based on 
    *        (cannot be 0 or 8)
    * @param offset offset into the 20 random bytes to draw random data from 
    *        (must be in range 0-16)
    * @param ch buffer for the challenge to be returned (must be of length 3 or more)
    *
    * @return <code>true</code> if successful, <code>false</code> if an error
    *         occurred  (use <code>getLastError()</code> for more
    *         information on the type of error)
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    *
    * @see #answerChallenge(byte[],byte[],byte[])
    * @see #getLastError()
    */
   public synchronized boolean generateChallenge (int page_number,
           int offset, byte[] ch)
      throws OneWireIOException, OneWireException
   {
      last_error = NO_ERROR;

      int addr = page_number << 5;

      if (ibc.eraseScratchPad(page_number))
      {
         ibc.useResume(true);

         if (ibc.SHAFunction(ibc.COMPUTE_CHALLENGE, addr))
         {
            int len = ibc.readScratchPad(scratchpad_buffer, 0);

            System.arraycopy(scratchpad_buffer, 8 + (offset % 17), ch, 0, 3);
            ibc.useResume(false);

            return true;
         }
         else
         {
            last_error = SHA_FUNCTION_ERROR;

            ibc.useResume(false);

            return false;
         }
      }

      last_error = ERASE_SCRATCHPAD_ERROR;

      ibc.useResume(false);

      return false;
   }

   private byte[] authentication_mac1            = new byte [20];
   private byte[] authentication_full_bind_code1 = new byte [15];
   private byte[] authentication_pad_challenge1  = new byte [32];
   private byte[] auth_challenge                 = new byte [3];

   /**
    * <p>Determines if the <code>SHAiButton</code> <code>user</code> belongs to the system
    * defined by this Coprocessor iButton.  This method can only be called by a 
    * Coprocessor iButton.  See the usage example in this class for initializing up
    * a Coprocessor iButton.</p>
    *
    * <p>The <code>SHAiButton</code> <code>user</code> must have previously been
    * initialized with the <code>setUser()</code> method.
    *
    * <p>The TMEX formatted page with the money signature is returned in 
    * the parameter <code>pagedata</code>.  <code>pagedata</code> must be
    * at least 32 bytes long to hold this data.  If the verification 
    * is successful, verify that the data in the returned page is valid 
    * with the <code>verifyUserMoney()</code> method.</p>
    *
    * <p>Failure of this method does not necessarily mean that 
    * the User iButton does not belong to the system.  It is possible that
    * a communication disruption here could cause a CRC error that 
    * would be indistinguishable from a failed authentication.  However,
    * repeated attempts should reveal whether it was truly a communication 
    * problem or a User iButton that does not belong to the system.</p>
    *
    * <p>This method starts by generating a random challenge on the
    * Coprocessor iButton.  The User iButton then answers that challenge
    * by signing a specified data page.  The Coprocessor then recreates
    * the User iButton's secret, and signs the same data that the User
    * signed.  If the signatures match, the authentication succeeded.</p>
    *
    * @param user the <code>SHAiButton</code> that will answer the 
    *        challenge and be authenticated, previously setup with <code>setUser()</code>
    * @param pagedata buffer for the data page that will be used by <code>user</code>
    *        to generate the authentication signature
    *
    * @return write cycle counter of the page used in authentication, or -1 if
    *        it could not authenticate (use <code>getLastError()</code> for more
    *        information on the type of error)
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    *
    * @see #generateChallenge(int,int,byte[])
    * @see #answerChallenge(byte[],byte[],byte[])
    * @see OneWireContainer18#bindSecretToiButton(int,byte[],byte[],int)
    * @see OneWireContainer18#SHAFunction(byte,int)
    * @see OneWireContainer18#matchScratchPad(byte[])
    * @see #verifyUserMoney(byte[],SHAiButton,int)
    * @see #getLastError()
    * @see #setUser(int)
    */
   public synchronized int verifyAuthentication (SHAiButton user,
           byte[] pagedata)
      throws OneWireException, OneWireIOException
   {
      last_error = NO_ERROR;

      //local cache-ing of stuff
      int    user_page                     = user.user_account_file_page;
      int    wspc_page_number              = workspace_page_number;
      byte[] authentication_pad_challenge  = authentication_pad_challenge1;
      byte[] authentication_full_bind_code = authentication_full_bind_code1;
      byte[] authentication_mac            = authentication_mac1;

      if (!isCoprocessor)
      {
         last_error = NO_COPROCESSOR_ERROR;

         return -1;
      }

      if (!user.isUser)
      {
         last_error = NO_USER_ERROR;

         return -1;
      }

      //steps here are
      //  1 generate a challenge
      //  2 make the user generate a response
      //  3 verify that response
      //NOTE! we can't get out of making the iButton generate the challenge here
      //because it messes with flags that alter the outcome of the SHA engine
      generateChallenge(authentication_page_number, 0, auth_challenge);

      //THIS IS OUR LONE COMMUNICATION WITH THE USER IBUTTON
      int write_cycle_counter = user.answerChallenge(auth_challenge,
                                   authentication_mac, pagedata);

      //AFTER THIS LETS PUT SOME OF THAT "RESUME" STUFF TO USE
      OneWireContainer18 ibcL = ibc;

      //now verify the response
      System.arraycopy(bind_code, 0, authentication_full_bind_code, 0, 4);

      authentication_full_bind_code [4] = ( byte ) user_page;

      System.arraycopy(user.address, 0, authentication_full_bind_code, 5, 7);
      System.arraycopy(bind_code, 4, authentication_full_bind_code, 12, 3);

      authentication_pad_challenge [8]  = ( byte ) (write_cycle_counter         & 0x0ff);
      authentication_pad_challenge [9]  = ( byte ) ((write_cycle_counter >> 8)  & 0x0ff);
      authentication_pad_challenge [10] = ( byte ) ((write_cycle_counter >> 16) & 0x0ff);
      authentication_pad_challenge [11] = ( byte ) ((write_cycle_counter >> 24) & 0x0ff);

      System.arraycopy(auth_challenge, 0, authentication_pad_challenge, 20, 3);

      //we've already copied user_account_file_page and iButton.address to full_bind_code,
      //now why don't we just do
      //      System.arraycopy(authentication_full_bind_code,4,authentication_pad_challenge,12,8);????
      //copies the iButton address and the page number from where it already is locally
      //skip that method invocation and extra assignment
      System.arraycopy(authentication_full_bind_code, 4,
                       authentication_pad_challenge, 12, 8);

      //the first thing bindSecretToiButton does with this thing is mask it with 7, so why do it twice?
      if (ibcL.bindSecretToiButton(authentication_page_number, bind_data,
                                   authentication_full_bind_code,
                                   wspc_page_number))
      {
         ibcL.useResume(true);

         if (ibcL.writeDataPage(wspc_page_number, pagedata))
         {
            if (ibcL.writeScratchPad(wspc_page_number, 0,
                                     authentication_pad_challenge, 0, 32))
            {
               int addr = wspc_page_number << 5;

               if (ibcL.SHAFunction(ibc.VALIDATE_DATA_PAGE, addr))
               {
                  if (ibcL.matchScratchPad(authentication_mac))
                  {
                     ibcL.useResume(false);

                     //pagedata has already been copied into where it needs to be
                     return write_cycle_counter;
                  }
                  else
                  {
                     last_error = AUTHENTICATION_FAILED_ERROR;
                  }
               }
               else
                  last_error = SHA_FUNCTION_ERROR;
            }
            else
               last_error = WRITE_SCRATCHPAD_ERROR;
         }
         else
            last_error = WRITE_MEMORY_PAGE_ERROR;
      }
      else
         last_error = BIND_SECRET_ERROR;

      ibc.useResume(false);

      return -1;
   }

   /**
    * <p>Determines if this <code>SHAiButton</codE> has enough of its parameters
    * set to act as a Coprocessor iButton.  To be a Coprocessor, the
    * following must be set:
    * <ul>
    *     <li><b>Coprocessor Service File Name and Extension</b><br>
    *         This is the name of the TMEX formatted file as it will appear
    *         on the User iButton.  When a User iButton is inserted, the host
    *         system must find this file and try to verify its data.
    *     </li>
    *     <li><b>Signing Page Number</b><br>
    *         This is the page the Coprocessor will use to create signatures that
    *         will be stored on User iButtons.  Only pages 0 and 8 can be used
    *         to create signatures on the DS1963S.  Since page 0 should be
    *         a directory file under TMEX, this should probably always be 
    *         page 8.
    *     </li>
    *     <li><b>Workspace Page Number</b><br>
    *         This is the page the Coprocessor will use to recreate the User 
    *         iButton secrets.  No data should be stored here, as it will be overwritten
    *         during authentication.
    *     </li>
    *     <li><b>Authentication Page Number</b><br>
    *         This is the page the Coprocessor's Authentication Secret is stored on.
    *         This secret will be used to recreate the User iButton specific secret.
    *     </li>
    * </ul></p>
    *
    * <p>The Coprocessor uses other parameters, but these have default and do not need
    * to be set unless required by the system.  These parameters are:
    * <ul>
    *     <li>Initial Signature: defaults to 20 bytes of 0's</li>
    *     <li>Signing Challenge: defaults to 3 bytes of 0's</li>
    *     <li>Binding Data: defaults to 32 0xff <code>byte</code>s</li>
    *     <li>Binding Code: defaults to 7 0xff <code>byte</code>s</li>
    * </ul>
    *
    * @return <code>true</code> if this iButton can act as a Coprocessor
    *
    * @see #setFilename(byte[],int)
    * @see #setSigningPageNumber(int)
    * @see #setAuthenticationPageNumber(int)
    * @see #setWorkspacePageNumber(int)
    * @see #setBindCode(byte[],int)
    * @see #setBindData(byte[],int)
    * @see #setSigningChallenge(byte[],int)
    * @see #setInitialSignature(byte[],int)
    */
   public boolean isCoprocessor ()
   {
      return isCoprocessor;
   }

   /**
    * Returns a <code>java.lang.String</code> with the unique 
    * address of the SHA iButton in hex.  This is only a convenience 
    * method that calls <code>OneWireContainer.getAddressAsString()</code>
    * on the <code>OneWireContainer18</code> object this <code>SHAiButton</code>
    * uses.
    *
    * @return unique address of the SHA iButton
    *
    * @see OneWireContainer#getAddressAsString()
    */
   public String toString ()
   {
      return ibc.getAddressAsString();
   }

   /**
    * <p>Sets up this <code>SHAiButton</code> as a User iButton.
    * To act as a User iButton, this object only needs to 
    * know what page number its service file is located in.
    * This is the page that contains signed data and/or 
    * is used for authentication.  Methods that take a User 
    * <code>SHAiButton</code> as a parameter check to make sure
    * that the User has been set up using this method.</p>
    *
    * <p>See the usage section for an example on how to discover the
    * page number where the service file is located on a User iButton.</p>
    *
    * @param file_page_number page number that contains signed data
    *
    * @return <code>true</code> if successful
    */
   public synchronized boolean setUser (int file_page_number)
   {
      if (file_page_number < 8)
      {
         file_page_number += 8;   //assume that it points to the first page of a multi-page file,

         //the next page must be the corresponding secure page
      }

      user_account_file_page = file_page_number;
      generic_money_page [0] = ( byte ) 0x1d;   //TMEX length

      //transaction ID is at locations 1-2            ]
      //money conversion factor is at locations 3-4   ]  this is all taken care
      //balance is at locations 5,6,7                 ]  of in the sign data function
      //signature is at locations 8-27                ]
      //data type is at location 28                   ]
      //continuation pointer is at location 29        ]
      //CRC is at locations 30,31                     ]
      //now the coprocessor must copy in initial_signature and balance, and then sign!
      isUser = true;

      return true;
   }

   /**
    * <p>Sets the optional initial signature for a Coprocessor iButton.  
    * When a data page is signed, the place in the page where the signature will 
    * be stored is initialized to this value.</p>
    * 
    * <p>Up to 20 bytes will be copied from <code>sig_ini</code> if
    * its length permits.  The initial signature defaults to 
    * all 0's.  If <code>sig_ini</code> is not long enough
    * to copy 20 bytes, the excess in the initial signature will
    * remain at all 0's.</p>
    * 
    * @param sig_ini new initial signature
    * @param start offset to start reading the new initial signature from <code>sig_ini</code>
    */
   public synchronized void setInitialSignature (byte[] sig_ini, int start)
   {
      System.arraycopy(sig_ini, start, initial_signature, 0,
                       ((sig_ini.length > 20 + start) ? 20
                                                      : sig_ini.length
                                                        - start));
   }

   /**
    * <p>Sets the optional signing challenge for a Coprocessor iButton.
    * All signatures on the Coprocessor have a location for
    * challenges.  This may optionally be set.  Note that this is
    * not the same as the challenge in the authentication sequence.
    * This is data that will be included with the Coprocessor's
    * signature of a data (money) file.</p>
    *
    * <p>Up to 3 bytes will be copied from <code>ch</code> if
    * its length permits.  The signing challenge defaults to 
    * all 0's.  If <code>ch</code> is not long enough
    * to copy 3 bytes, the excess in the signing challenge will
    * remain at all 0's.</p>
    * 
    * @param ch Byte new random challenge
    * @param start offset to start reading the new signing challenge from <code>ch</code>
    */
   public synchronized void setSigningChallenge (byte[] ch, int start)
   {
      System.arraycopy(ch, start, signing_challenge, 0,
                       ((ch.length > 3 + start) ? 3
                                                : ch.length - start));
   }

   /**
    * <p>Sets the file name and extension for the Coprocessor 
    * Service File.  This file will appear on User iButtons
    * and will contain data signed by the Coprocessor.</p>
    *
    * <p>The parameter <code>buf</code> should have 5 bytes
    * to represent the Coprocessor Service File name and 
    * extension.  For example, if the service file is named
    * "DLSM.102" (extension = 102), <code>buf</code> should look 
    * like:
    * <code><pre>
    *   buf[start + 0] = (byte)'D';
    *   buf[start + 1] = (byte)'L';
    *   buf[start + 2] = (byte)'S';
    *   buf[start + 3] = (byte)'M';
    *   buf[start + 4] = (byte)102;
    * </pre></code></p>
    *
    * <p>This function <em>must</em> be called to initialize
    * a Coprocessor iButton.</p>
    *
    * @param buf the file name and extension
    * @param start index to start copying from <code>buf</code>
    *
    * @see #isCoprocessor()
    * @see #getUserFileName()
    * @see #getUserFileExtension()
    */
   public synchronized void setFilename (byte[] buf, int start)
   {
      filename = new byte [5];

      System.arraycopy(buf, start, filename, 0, 5);
      checkCoprocessor();
   }

   /**
    * <p>Sets the page used to generate signatures on the Coprocessor
    * iButton.  The hardware of the DS1963S allows only page
    * 0 or page 8 to be used for this function.  Since this class
    * assumes the TMEX file structure (requiring page 0 to be 
    * a directory page), this method should be called setting the
    * signing page to page 8.</p>
    * 
    * <p>This function <em>must</em> be called to initialize
    * a Coprocessor iButton.</p>
    *
    * @param pg page number for generating signatures
    *
    * @see #isCoprocessor()
    */
   public synchronized void setSigningPageNumber (int pg)
   {
      if (pg == 0)
         pg = 8;

      signing_page_number = pg;

      checkCoprocessor();
   }

   /**
    * <p>Sets the page that the authentication secret is
    * installed on in this Coprocessor iButton.  This secret
    * will be used to recreate User iButton specific authentication
    * secrets during the authentication process.</p>
    * 
    * <p>This function <em>must</em> be called to initialize
    * a Coprocessor iButton.</p>
    *
    * @param pg page the installation secret is installed on
    *
    * @see #isCoprocessor()
    */
   public synchronized void setAuthenticationPageNumber (int pg)
   {
      if (pg < 8)
         pg += 8;

      authentication_page_number = pg;

      checkCoprocessor();
   }

   /**
    * <p>Sets the page this Coprocessor iButton will use to recreate
    * User iButton specific secrets.  No data should be stored here, as it
    * will be overridden when the system tries to perform an
    * authentication.</p>
    * 
    * <p>This function <em>must</em> be called to initialize
    * a Coprocessor iButton.</p>
    *
    * @param pg free page that can be used as workspace
    *
    * @see #isCoprocessor()
    */
   public synchronized void setWorkspacePageNumber (int pg)
   {
      workspace_page_number = pg;

      checkCoprocessor();
   }


   //see if all the parameters for a coprocessor have been set
   private boolean checkCoprocessor ()
   {
      if ((signing_page_number != -1) && (authentication_page_number != -1)
              && (workspace_page_number != -1) && (filename != null))
         isCoprocessor = true;

      return isCoprocessor;
   }

   /**
    * Returns the name of the Coprocessor Service 
    * File.  Throws a <code>java.lang.NullPointerException</code>
    * if the file name has not been set.  If the file name
    * is 'DLSM.102', this returns the <code>java.lang.String</code>
    * "DLSM".
    *
    * @return the name of the Coprocessor Service File
    *
    * @see #setFilename(byte[],int)
    */
   public String getUserFileName ()
   {

      //return user_account_file;
      return new String(filename, 0, 4);
   }

   /**
    * Returns the extension of the Coprocessor Service 
    * File.  Throws a <code>java.lang.NullPointerException</code>
    * if the file name has not been set.  If the file name
    * is 'DLSM.102', this returns the <code>int</code>
    * 102.
    *
    * @return the extension of the Coprocessor Service File
    *
    * @see #setFilename(byte[],int)
    */
   public int getUserFileExtension ()
   {

      //return user_account_file_extension;
      return (filename [4] & 0x0ff);
   }

   private byte[] sign_scratchpad1 = new byte [32];

   /**
    * <p>Makes this Coprocessor iButton produce a signature 
    * for a new amount of money for the User iButton, then places 
    * the new money file on <code>user</code>.</p>
    *
    * <p>The write cycle counter is part of a valid signature.
    * The Coprocessor iButton uses it to verify a signature
    * of a signed data file.  It should be returned in a normal 
    * transaction from the method <code>verifyAuthentication()</code>.
    * However, if this method fails or if this is the 
    * initialization of a User iButton, use -1 for the write cycle counter.
    * Then, the code will determine the correct write cycle counter 
    * value directly from the User iButton.  However, do not do this
    * in a normal transaction sequence as it takes longer.</p>
    *
    * <p>Note that this method specifically handles TMEX file formats.</p>
    *
    * @param user <code>SHAiButton</code> object that has been initialized as a user
    *        with <code>setUser(int)</code>
    * @param newbalance new balance to be stored on the User iButton
    * @param write_cycle_counter the current write cycle counter of the User iButton's
    *        money page
    * @param pagedata the money data page (from <code>verifyAuthentication()</code>)
    *
    * @return <code>true</code> if successful, <code>false</code>
    *         if an error occurred (use <code>getLastError()</code> for more
    *         information on the type of error)
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    *
    * @see #verifyUserMoney(byte[],SHAiButton,int)
    * @see #getLastError()
    * @see #verifyAuthentication(SHAiButton,byte[])
    */
   public synchronized boolean signDataFile (SHAiButton user, int newbalance,
                                             int write_cycle_counter,
                                             byte[] pagedata)
      throws OneWireIOException, OneWireException
   {
      last_error = NO_ERROR;

      if (!isCoprocessor)
      {
         last_error = NO_COPROCESSOR_ERROR;

         return false;
      }

      if (!user.isUser)
      {
         last_error = NO_USER_ERROR;

         return false;
      }

      //local cache
      byte[] user_generic_money_page     = user.generic_money_page;
      int    user_user_account_file_page = user.user_account_file_page;
      byte[] sign_scratchpad             = sign_scratchpad1;

      System.arraycopy(pagedata, 1, user_generic_money_page, 1, 4);

      user_generic_money_page [28] = pagedata [28];

      System.arraycopy(initial_signature, 0, user_generic_money_page, 8,
                       initial_signature.length > 20 ? 20
                                                     : initial_signature.length);

      user_generic_money_page [5]  = ( byte ) newbalance;
      user_generic_money_page [6]  = ( byte ) (newbalance >> 8);
      user_generic_money_page [7]  = ( byte ) (newbalance >> 16);
      user_generic_money_page [30] = ( byte ) 0;
      user_generic_money_page [31] = ( byte ) 0;

      int wcc = 0;

      if (write_cycle_counter == -1)
      {

         //this should only be done on initialization of user ibutton by coprocessor,
         //not during transactions
         //now we also need to get things like wcc, user_page_number, user ID
         byte[] wcc_page = new byte [32];   //this is OK, we are not in critical path here

         //so malloc-ing doesnt really matter
         user.ibc.readMemoryPage(19, wcc_page, 0);

         //we also need to increment the WCC since we are going to write then later verify
         int offset = (user_user_account_file_page & 7) << 2;

         wcc = wcc_page [offset++] & 0x0ff;
         wcc = wcc | ((wcc_page [offset++] & 0x0ff) << 8);
         wcc = wcc | ((wcc_page [offset++] & 0x0ff) << 16);
         wcc = wcc | ((wcc_page [offset] & 0x0ff) << 24);
      }
      else
         wcc = write_cycle_counter;

      //we need to increment the write_cycle_counter since we will be writing to the part
      wcc++;

      sign_scratchpad [8]  = ( byte ) (wcc & 0x0ff);
      sign_scratchpad [9]  = ( byte ) ((wcc >> 8) & 0x0ff);
      sign_scratchpad [10] = ( byte ) ((wcc >> 16) & 0x0ff);
      sign_scratchpad [11] = ( byte ) ((wcc >> 24) & 0x0ff);
      sign_scratchpad [12] = ( byte ) user_user_account_file_page;

      System.arraycopy(user.address, 0, sign_scratchpad, 13, 7);
      System.arraycopy(signing_challenge, 0, sign_scratchpad, 20, 3);

      OneWireContainer18 ibcL = ibc;

      //now we are ready to make a signature
      if (!ibcL.writeDataPage(signing_page_number, user_generic_money_page))   //writing to THIS ibutton
      {
         last_error = WRITE_MEMORY_PAGE_ERROR;

         return false;
      }

      ibcL.useResume(true);

      if (!ibcL.writeScratchPad(0, 0, sign_scratchpad, 0, 32))
      {
         last_error = WRITE_SCRATCHPAD_ERROR;

         ibcL.useResume(false);

         return false;
      }

      //sign that baby!
      int addr = signing_page_number << 5;

      if (!ibcL.SHAFunction(ibc.SIGN_DATA_PAGE, addr))
      {
         last_error = SHA_FUNCTION_ERROR;

         ibcL.useResume(false);

         return false;
      }

      //after signature make sure to dump in the inverted CRC
      ibcL.readScratchPad(sign_scratchpad, 0);
      System.arraycopy(sign_scratchpad, 8, user.generic_money_page, 8, 20);

      int crc = ~CRC16.compute(user_generic_money_page, 0,
                               user_generic_money_page [0] + 1,
                               user_user_account_file_page);

      user_generic_money_page [30] = ( byte ) crc;
      user_generic_money_page [31] = ( byte ) (crc >> 8);

      ibcL.useResume(false);

      //now put the page on the user ibutton
      if (!user.ibc.writeDataPage(user_user_account_file_page,
                                  user_generic_money_page))
      {
         last_error = WRITE_MEMORY_PAGE_ERROR;

         return false;
      }

      return true;
   }

   private byte[] verify_macpage     = new byte [20];
   private byte[] verify_scratchpad1 = new byte [32];

   /**
    * Verifies a User iButton's signed data on this Coprocessor iButton.
    * The Coprocessor must recreate the signature based on the data in the 
    * file and the write cycle counter of the file, and then match that
    * with the signature embedded in the file from the User iButton.
    *
    * @param userpage the full 32 byte TMEX file from the User iButton
    *        (from <code>verifyAuthentication()</code>)
    * @param user <code>SHAiButton</code> object that has been initialized as a user
    *        with <code>setUser(int)</code>
    * @param wcc write cycle counter of the user's money page 
    *        (from <code>verifyAuthentication()</code>)
    *
    * @return <code>true<code> if the data file is valid, <code>false</code>
    *         if an error occurred (use <code>getLastError()</code> for more
    *         information on the type of error)
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    *
    * @see #verifyAuthentication(SHAiButton,byte[])
    * @see #getLastError()
    */
   public synchronized boolean verifyUserMoney (byte[] userpage,
           SHAiButton user, int wcc)
      throws OneWireIOException, OneWireException
   {

      //WE TALK EXCLUSIVELY TO THE COPROCESSOR HERE SO WE CAN USE RESUME
      //  trick - call writeDataPage first then set resume to true.  writeDataPage only does one select
      //but then we don't have an extra method invocation, so its down below
      last_error = NO_ERROR;

      int    page              = user.user_account_file_page;
      byte[] verify_scratchpad = verify_scratchpad1;

      if (!isCoprocessor)
      {
         last_error = NO_COPROCESSOR_ERROR;

         return false;
      }

      if (!user.isUser)   
      {
         last_error = NO_USER_ERROR;

         return false;
      }

      System.arraycopy(userpage, 8, verify_macpage, 0, 20);

      //now lets 0-out the mac and the CRC
      System.arraycopy(initial_signature, 0, userpage, 8, 20);

      userpage [30] = ( byte ) 0;
      userpage [31] = ( byte ) 0;

      //now we also need to get things like wcc, user_page_number, user ID
      verify_scratchpad [8]  = ( byte ) (wcc);
      verify_scratchpad [9]  = ( byte ) (wcc >> 8);
      verify_scratchpad [10] = ( byte ) (wcc >> 16);
      verify_scratchpad [11] = ( byte ) (wcc >> 24);
      verify_scratchpad [12] = ( byte ) page;

      System.arraycopy(user.address, 0, verify_scratchpad, 13, 7);
      System.arraycopy(signing_challenge, 0, verify_scratchpad, 20, 3);

      //now we are ready to make a signature
      if (!ibc.writeDataPage(signing_page_number, userpage))   //writing to THIS ibutton
      {
         last_error = WRITE_MEMORY_PAGE_ERROR;

         ibc.useResume(false);

         return false;
      }

      ibc.useResume(true);

      if (!ibc.writeScratchPad(0, 0, verify_scratchpad, 0, 32))
      {
         last_error = WRITE_SCRATCHPAD_ERROR;

         ibc.useResume(false);

         return false;
      }

      //sign that baby!
      int addr = signing_page_number << 5;

      if (!ibc.SHAFunction(ibc.VALIDATE_DATA_PAGE, addr))
      {
         last_error = SHA_FUNCTION_ERROR;

         ibc.useResume(false);

         return false;
      }
      else
      {
         if (ibc.matchScratchPad(verify_macpage))
         {
            ibc.useResume(false);

            return true;
         }
         else
         {
            last_error = VERIFICATION_FAILED_ERROR;
         }
      }

      ibc.useResume(false);

      return false;
   }

   private byte[] pg0 = new byte [32];

   /**
    * <p>Reads a TMEX file starting at the specified page.  The parameter
    * <code>page</code> will store the file, or up to <code>page.length</code>
    * bytes of the file.  This method performs CRC checks to ensure
    * the validity of the file.</p>
    * 
    * <p>Call <code>readFile(0)</code> to read the root directory.</p>
    *
    * @param start_page page of the DS1963S to start reading from
    * @param page buffer for the file data
    *
    * @return length of data read
    *
    * @throws OneWireIOException on a 1-Wire communication error such as 
    *         reading an incorrect CRC from a 1-Wire device.  This could be
    *         caused by a physical interruption in the 1-Wire Network due to 
    *         shorts or a newly arriving 1-Wire device issuing a 'presence pulse'.
    * @throws OneWireException on a communication or setup error with the 1-Wire 
    *         adapter
    */
   public synchronized int readFile (int start_page, byte[] page)
      throws OneWireIOException, OneWireException
   {
      last_error = NO_ERROR;

      int index           = 0;
      int cont_ptr        = -1;
      int length_to_write = 0;

      while ((cont_ptr != 0) && (index < page.length))
      {
         ibc.readMemoryPage(start_page, pg0, 0);

         int len = pg0 [0];

         if ((0x0ff & len) > 32)   //bad read, not a file...
            return 0;

         length_to_write = (index + len - 1 > page.length)
                           ? page.length - index
                           : len - 1;

         //baos.write(pg0,1,len-1);
         System.arraycopy(pg0, 1, page, index, length_to_write);

         index    += length_to_write;
         cont_ptr = pg0 [len];

         if (CRC16.compute(pg0, 0, len + 3, start_page) != 0xb001)
         {
            last_error = CRC_ERROR;

            return 0;
         }

         start_page = cont_ptr;
      }

      return index;
   }

   /**
    * Returns the last error that occurred.  Most errors
    * indicate a communication or setup problem, though some
    * errors, if repeated, can indicate that an iButton 
    * does not belong to the system (cannot authenticate)
    * or has an invalid money file (cannot verify signed data).
    *
    * @return the last error
    *
    * @see #NO_ERROR
    * @see #ERASE_SCRATCHPAD_ERROR
    * @see #WRITE_SCRATCHPAD_ERROR
    * @see #READ_SCRATCHPAD_ERROR
    * @see #READ_MEMORY_PAGE_ERROR
    * @see #READ_AUTHENTICATED_ERROR
    * @see #BIND_SECRET_ERROR
    * @see #WRITE_MEMORY_PAGE_ERROR
    * @see #SHA_FUNCTION_ERROR
    * @see #CRC_ERROR
    * @see #NO_COPROCESSOR_ERROR
    * @see #NO_USER_ERROR
    * @see #AUTHENTICATION_FAILED_ERROR
    * @see #VERIFICATION_FAILED_ERROR
    */
   public int getLastError ()
   {
      return last_error;
   }
}
