
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.onewire.utils;

import com.dalsemi.onewire.adapter.*;
import com.dalsemi.onewire.container.*;

/**
 * This class encapsulates a 1-Wire network event such as
 * arrival or departure.  This class is created by an instance of <code>OneWireMonitor</code>,
 * and passed into a class implementing <code>OneWireMonitorEventListener</code> using
 * the methods <code>oneWireArrival()</code> and <code>oneWireDeparture()</code>
 *
 *
 * @see OneWireMonitor
 * @see OneWireMonitorEventListener
 */
public class OneWireMonitorEvent
   extends java.util.EventObject
{
   private DSPortAdapter adapter;
   private long          address;

   /**
    * Create a new OneWireMonitorEvent with specified adapter, and address
    *
    * @param owm reference to <code>OneWireMonitor</code> generating the event
    * @param adapter 1-Wire adapter where device is located
    * @param address 1-Wire network address of device
    *
    */
   public OneWireMonitorEvent (OneWireMonitor owm, DSPortAdapter adapter,
                               long address)
   {
      // Set the source of the event in the superclass
      super(owm);

      this.adapter = adapter;
      this.address = address;
   }

   /**
    * Retrieves the container for the device causing the event.
    *
    * @return <code>OneWireContainer</code> encapsulating the state of the
    * 1-Wire device generating the event
    */
   public OneWireContainer getDeviceContainer ()
   {
      return getAdapter().getDeviceContainer(address);
   }

   /**
    * Retrieves the adapter on which the event occurred.
    *
    * @return <code>DSPortAdapter</code> of the adapter on which the 1-Wire event was generated
    */
   public DSPortAdapter getAdapter ()
   {

      // FIXME: Need to clone the adapter before we give someone else control.
      // return adapter.clone();
      return adapter;
   }

   /**
    * Retrieves the address of the 1-Wire device causing the event.
    *
    * @return The address of the 1-Wire device
    * @see com.dalsemi.onewire.utils.Address
    */
   public byte[] getAddress ()
   {
      return Address.toByteArray(address);
   }

   /**
    * Retrieves the address of the 1-Wire device causing the event.
    *
    * @return The address of the 1-Wire device
    * @see com.dalsemi.onewire.utils.Address
    */
   public long getAddressAsLong ()
   {
      return address;
   }

   /**
    * Retrieves the address of the 1-Wire device causing the event.
    *
    * @return The address of the 1-Wire device
    * @see com.dalsemi.onewire.utils.Address
    */
   public String getAddressAsString ()
   {
      return Address.toString(address);
   }
}
